/* 
 * RC-like remote control over UDP/IPv4; receiver part.
 * Copyright (c) 2005, 2006 pascal@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

void fatal(char *msg) {
  if ( errno ) perror(msg);
  else fprintf(stderr, "%s\n", msg);
  exit(1);
}

#define NAME_LENGTH 128

#define MAX_CHANS 6

int main (int argc, char **argv) {
  unsigned short port = 9001;
  int sock;
  unsigned short port2 = 0;
  int sock2 = -1;
  char *dev = "/dev/opwm0";
  int fd;
  int timeout = 1000;
  int verbose = 0;

  for ( int i=1; i<argc; ++i ) {
    if      ( !strcmp(argv[i], "-c") && ++i<argc ) dev = argv[i];
    else if ( !strcmp(argv[i], "-p") && ++i<argc ) port = atoi(argv[i]);
    else if ( !strcmp(argv[i], "-p2") && ++i<argc ) port2 = atoi(argv[i]);
    else if ( !strcmp(argv[i], "-t") && ++i<argc ) timeout = atoi(argv[i]);
    else if ( !strcmp(argv[i], "-v") ) ++verbose;
    else fatal("Usage:");
  }
  
  fd = open(dev, O_WRONLY);
  if ( fd < 0 ) fatal(dev);

  {
    sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if ( sock < 0 ) fatal("socket");
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    addr.sin_port = htons(port);
    if ( bind(sock, (struct sockaddr*)&addr, sizeof(addr)) < 0 )
      fatal("bind");
  }

  if ( port2 ) {
    sock2 = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if ( sock2 < 0 ) fatal("socket2");
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    addr.sin_port = htons(port2);
    if ( bind(sock2, (struct sockaddr*)&addr, sizeof(addr)) < 0 )
      fatal("bind2");
    if ( fcntl(sock2, O_NONBLOCK, 1) < 0 ) fatal("fcntl2");
  }

  int16_t chans2[MAX_CHANS] = { 1, };
  fd_set fds;
  FD_ZERO(&fds);
    
  while ( 1 ) {
    struct msg_t {
      uint32_t flags;
      uint16_t chans[MAX_CHANS];
    } msg;
    struct timeval tv;
    tv.tv_sec = timeout / 1000;
    tv.tv_usec = (timeout % 1000) * 1000;
    FD_SET(sock, &fds);
    if ( select(sock+1, &fds, NULL, NULL, &tv) <= 0 )
      fatal("receive timeout");
    int nr = recvfrom(sock, &msg, sizeof(msg), 0, NULL, NULL);
    if ( nr < 0 ) fatal("recvfrom");
    if ( nr != sizeof(msg) ) fatal("msg format");
    for ( int i=0; i<MAX_CHANS; ++i ) msg.chans[i] = ntohs(msg.chans[i]);
    msg.flags = ntohl(msg.flags);
    if ( sock2 >= 0 ) {
      do {
	nr = recvfrom(sock2, chans2, sizeof(chans2), 0, NULL, NULL);
      } while ( nr > 0 );
      if ( msg.flags & 1 )
	for ( int i=0; i<MAX_CHANS; ++i ) msg.chans[i] += ntohs(chans2[i]);
    }	
    if ( write(fd, msg.chans, sizeof(msg.chans)) < 0 ) fatal("write");
    if ( verbose ) {
      int i;
      printf("%08x", msg.flags);
      for ( i=0; i<nr; ++i ) printf(" %d", msg.chans[i]);
      printf("\n");
      fflush(stdout);
    }
  }

  return 0;
}
